import numpy as np

from ..base import Base


class mode(Base):
    def calc(self, dset: str, f: float, c: int):
        if f"modes/{dset}/arr" not in self.llyr.dsets:
            print("Calculating modes ...")
            self.llyr.calculate_modes(dset)
        fi = int((np.abs(self.llyr[f"modes/{dset}/freqs"][:] - f)).argmin())
        _mode = self.llyr[f"modes/{dset}/arr"][fi][..., c]
        return _mode



from typing import Optional

import numpy as np
import dask.array as da
import h5py

from ..base import Base


class disp(Base):
    def calc(
        self,
        dset: str,
        name: Optional[str] = None,
        override: Optional[bool] = False,
        tslice=slice(None),
        zslice=slice(None),
        yslice=slice(None),
        xslice=slice(None),
        cslice=2,
    ):
        with h5py.File(self.llyr.path, "r") as f:
            arr = da.from_array(f[dset], chunks=(None, None, 16, None, None))
            arr = arr[(tslice, zslice, yslice, xslice, cslice)]  # slice
            arr = da.multiply(
                arr, np.hanning(arr.shape[0])[:, None, None, None]
            )  # hann filter on the t axis
            arr = arr.sum(axis=1)  # t,z,y,x => t,y,x sum of z
            arr = da.moveaxis(arr, 1, 0)  # t,y,x => y,t,x swap t and y
            ham2d = np.sqrt(
                np.outer(np.hanning(arr.shape[1]), np.hanning(arr.shape[2]))
            )  # shape(t, x)
            arr = da.multiply(arr, ham2d[None, :, :])  # hann window on t and x
            arr = da.fft.fft2(arr)  # 2d fft on t and x
            arr = da.subtract(
                arr, da.average(arr, axis=(1, 2))[:, None, None]
            )  # substract the avr of t,x for a given y
            arr = da.moveaxis(arr, 0, 1)
            arr = arr[: arr.shape[0] // 2]  # split f in 2, take 1st half
            arr = da.fft.fftshift(arr, axes=(1, 2))
            arr = da.absolute(arr)  # from complex to real
            arr = da.sum(arr, axis=1)  # sum y
            arr = arr.compute()

        freqs = np.fft.rfftfreq(arr.shape[0], self.llyr.dt)
        kvecs = np.fft.fftshift(np.fft.fftfreq(arr.shape[1], self.llyr.dx)) * 1e-6
        if name is not None:
            self.llyr.add_dset(arr, f"{name}/arr", override)
            self.llyr.add_dset(freqs, f"{name}/freqs", override)
            self.llyr.add_dset(kvecs, f"{name}/kvecs", override)

        return arr



import matplotlib.pyplot as plt
import numpy as np

from ..base import Base


class modes(Base):
    def plot(self, dset: str, f: float, axes=None):
        mode1 = self.llyr.calc.mode(dset, f, 0)
        mode2 = self.llyr.calc.mode(dset, f, 1)
        mode3 = self.llyr.calc.mode(dset, f, 2)
        mode_list = np.array([mode1, mode2, mode3])
        mode_list_max = np.abs(mode_list).max()
        extent = [
            0,
            mode1.shape[1] * self.llyr.dy * 1e9,
            0,
            mode1.shape[0] * self.llyr.dx * 1e9,
        ]

        if axes is None:
            fig, axes = plt.subplots(3, 3, sharex=True, sharey=True, figsize=(4, 4))
        else:
            fig = axes[0, 0].figure
        for c in range(3):
            mode_abs = np.abs(mode_list[c])
            mode_ang = np.angle(mode_list[c])
            alphas = mode_abs / mode_abs.max()

            axes[0, c].imshow(
                mode_abs,
                cmap="inferno",
                vmin=0,
                vmax=mode_list_max,
                extent=extent,
                aspect="equal",
            )
            axes[1, c].imshow(
                mode_ang,
                aspect="equal",
                cmap="hsv",
                vmin=-np.pi,
                vmax=np.pi,
                interpolation="None",
                extent=extent,
            )
            axes[2, c].imshow(
                mode_ang,
                aspect="equal",
                alpha=alphas,
                cmap="hsv",
                vmin=-np.pi,
                vmax=np.pi,
                interpolation="None",
                extent=extent,
            )
        # axes[0, 0].set_title(r"$m_x$")
        # axes[0, 1].set_title(r"$m_y$")
        # axes[0, 2].set_title(r"$m_z$")
        # axes[0, 0].set_ylabel(r"$y$ (nm)")
        # axes[1, 0].set_ylabel(r"$y$ (nm)")
        # axes[2, 0].set_ylabel(r"$y$ (nm)")
        # axes[2, 0].set_xlabel(r"$x$ (nm)")
        # axes[2, 1].set_xlabel(r"$x$ (nm)")
        # axes[2, 2].set_xlabel(r"$x$ (nm)")
        # cb = fig.colorbar(
        #     axes[0, 2].get_images()[0], cax=axes[0, 2].inset_axes((1.05, 0.0, 0.05, 1))
        # )
        # cb.ax.set_ylabel("Amplitude")
        # for i in [1, 2]:
        #     cb = fig.colorbar(
        #         axes[1, 2].get_images()[0],
        #         cax=axes[i, 2].inset_axes((1.05, 0.0, 0.05, 1)),
        #         ticks=[-3, 0, 3],
        #     )
        #     cb.set_ticklabels([r"-$\pi$", 0, r"$\pi$"])
        #     cb.ax.set_ylabel("Phase")
        # fi = (np.abs(self.llyr[f"mode_list/{dset}/freqs"][:] - f)).argmin()
        # ff = self.llyr[f"mode_list/{dset}/freqs"][:][fi]
        fig.suptitle(f"{self.llyr.name}")
        # fig.tight_layout()
        for ax in axes.flatten():
            ax.set(xticks=[], yticks=[])
        return self



from typing import Optional, Tuple, Union, Dict
import os
from pathlib import Path

import numpy as np
import h5py

from .plot import Plot
from .calc import Calc
from .h5 import H5
from .make import Make
from .ovf import save_ovf
from ._utils import cspectra_b


class Llyr:
    def __init__(self, path: str) -> None:
        path = os.path.abspath(path)
        self._add_path(path)
        self._getitem_dset: Optional[str] = None
        self.plot = Plot(self)
        self.calc = Calc(self)
        self.h5 = H5(self)
        self.make = Make(self).make

    def _add_path(self, path):
        p = Path(path)
        if p.suffix == "":
            self.name = p.name
        else:
            self.name = p.name.replace(p.suffix, "")
        self.path = f"{str(p.parent)}/{self.name}.h5"

    def __repr__(self) -> str:
        return f"Llyr('{self.name}')"

    def __str__(self) -> str:
        return f"Llyr('{self.name}')"

    def __getitem__(
        self,
        index: Union[str, Tuple[Union[int, slice], ...]],
    ) -> Union["Llyr", float, np.ndarray]:
        if isinstance(index, (slice, tuple, int)):
            # if dset is defined
            if isinstance(self._getitem_dset, str):
                out_dset: np.ndarray = self.h5.get_dset(self._getitem_dset, index)
                self._getitem_dset = None
                return out_dset
            else:
                raise AttributeError("You can only slice datasets")

        elif isinstance(index, str):
            # if dataset
            if index in self.dsets:
                self._getitem_dset = index
                return self
            # if attribute
            elif index in self.attrs:
                out_attribute: float = self.attrs[index]
                return out_attribute
            else:
                raise KeyError("No such Dataset or Attribute")
        else:
            print()
            raise TypeError(f"{index}, {type(index)}")

    @property
    def mx3(self) -> None:
        print(self["mx3"])

    @property
    def dt(self) -> float:
        return self.attrs["dt"]

    @property
    def dx(self) -> float:
        return self.attrs["dx"]

    @property
    def dy(self) -> float:
        return self.attrs["dy"]

    @property
    def dz(self) -> float:
        return self.attrs["dz"]

    @property
    def pp(self) -> None:
        print("Datasets:")
        for dset_name, dset_shape in self.dsets.items():
            print(f"    {dset_name:<20}: {dset_shape}")
        print("Global Attributes:")
        for key, val in self.attrs.items():
            if key in ["mx3", "script", "logs"]:
                val = val.replace("\n", "")
                print(f"    {key:<20}= {val[:10]}...")
            else:
                print(f"    {key:<20}= {val}")

    @property
    def dsets(self) -> Dict[str, str]:
        def add_dset(name, obj):
            # pylint: disable=protected-access
            if isinstance(obj, h5py._hl.dataset.Dataset):
                dsets[name] = obj.shape

        dsets: Dict[str, str] = {}
        with h5py.File(self.path, "r") as f:
            f.visititems(add_dset)
        return dsets

    @property
    def attrs(self) -> dict:
        attrs = {}
        with h5py.File(self.path, "r") as f:
            for k, v in f.attrs.items():
                attrs[k] = v
        return attrs

    @property
    def snap(self):
        self.plot.snapshot_png("stable")


cspectra = cspectra_b(Llyr)



from .disp import disp
from .fft_tb import fft_tb
from .fft import fft
from .mode import mode
from .sk_number import sk_number


class Calc:
    def __init__(self, llyr):
        self.disp = disp(llyr).calc
        self.fft_tb = fft_tb(llyr).calc
        self.fft = fft(llyr).calc
        self.mode = mode(llyr).calc
        self.sk_number = sk_number(llyr).calc



from typing import Tuple, Union, Optional
import os
import multiprocessing as mp

import numpy as np
from tqdm import tqdm
import h5py

from .ovf import save_ovf, load_ovf


class H5:
    def __init__(self, llyr):
        self.llyr = llyr

    def save_as_ovf(self, arr: np.ndarray, name: str):
        path = self.llyr.path.replace(f"{self.llyr.name}", f"{name}.ovf")
        save_ovf(path, arr, self.llyr.dx, self.llyr.dy, self.llyr.dz)

    def create_h5(self, override: bool) -> bool:
        """Creates an empty .h5 file"""
        if override:
            with h5py.File(self.llyr.path, "w"):
                return True
        else:
            if os.path.isfile(self.llyr.path):
                input_string: str = input(
                    f"{self.llyr.path} already exists, overwrite it [y/n]?"
                )
                if input_string.lower() in ["y", "yes"]:
                    with h5py.File(self.llyr.path, "w"):
                        return True
        return False

    def shape(self, dset: str) -> Tuple:
        with h5py.File(self.llyr.path, "r") as f:
            return f[dset].shape

    def delete(self, dset: str) -> None:
        """deletes dataset"""
        with h5py.File(self.llyr.path, "a") as f:
            del f[dset]

    def move(self, source: str, destination: str) -> None:
        """move dataset or attribute"""
        with h5py.File(self.llyr.path, "a") as f:
            f.move(source, destination)

    def add_attr(
        self,
        key: str,
        val: Union[str, int, float, slice, Tuple[Union[int, slice], ...]],
        dset: Optional[str] = None,
    ) -> None:
        """set a new attribute"""
        if dset is None:
            with h5py.File(self.llyr.path, "a") as f:
                f.attrs[key] = val
        else:
            with h5py.File(self.llyr.path, "a") as f:
                f[dset].attrs[key] = val

    def add_dset(self, arr: np.ndarray, name: str, override: bool = False):
        if name in self.llyr.dsets:
            if override:
                self.delete(name)
            else:
                raise NameError(
                    f"Dataset with name '{name}' already exists, you can use 'override=True'"
                )
        with h5py.File(self.llyr.path, "a") as f:
            f.create_dataset(name, data=arr)

    def get_dset(self, dset, slices):
        with h5py.File(self.llyr.path, "r") as f:
            return f[dset][slices]

    def get_attrs(self, dset):
        with h5py.File(self.llyr.path, "r") as f:
            return dict(f[dset].attrs)

    def load_dset(self, name: str, dset_shape: tuple, ovf_paths: list) -> None:
        with h5py.File(self.llyr.path, "a") as f:
            dset = f.create_dataset(name, dset_shape, np.float32)
            with mp.Pool(processes=int(mp.cpu_count())) as p:
                for i, data in enumerate(
                    tqdm(
                        p.imap(load_ovf, ovf_paths),
                        leave=False,
                        desc=name,
                        total=len(ovf_paths),
                    )
                ):
                    dset[i] = data



import numpy as np

from ..base import Base


class fft_tb(Base):
    def calc(
        self,
        dset: str,
        tmax: int = None,
        tmin: int = None,
        tstep: int = 1,
        normalize: bool = False,
    ):
        if self.llyr.h5.shape("table/t")[0] > 2000:
            tstep = 20
        y = self.llyr[f"table/{dset}"][slice(tmin, tmax, tstep)]
        ts = self.llyr["table/t"][:]
        table_dt = (ts[-1] - ts[0]) / len(ts)
        x = np.fft.rfftfreq(y.shape[0], table_dt * tstep) * 1e-9
        y -= y[0]
        y -= np.average(y)
        y = np.multiply(y, np.hanning(y.shape[0]))
        y = np.fft.rfft(y)
        y = np.abs(y)
        if normalize:
            y /= y.max()
        return x, y



import matplotlib.pyplot as plt
import peakutils

from ..base import Base


class fft_tb(Base):
    def plot(
        self,
        tmin=0,
        tmax=-1,
        fft_tmin=0,
        fft_tmax=-1,
        tstep=1,
        thres=0.15,
        axes=None,
    ):
        if axes is None:
            _, axes = plt.subplots(3, 1, sharex=True)
        comps = ["mx", "my", "mz"]
        fss = []
        for i in range(3):
            c, ax = comps[i], axes[i]
            x, y = self.llyr.calc.fft_tb(c, tmax=fft_tmax, tmin=fft_tmin, tstep=tstep)
            ax.plot(x[tmin:tmax], y[tmin:tmax])
            list_peaks = peakutils.indexes(y, thres=thres, min_dist=4)
            fs = [x[i] for i in list_peaks]
            fss.append(fs)
            for f in fs:
                ax.axvline(f, ls="--", c="gray")
                ax.text(
                    f,
                    y[tmin:].max() * 1.15,
                    f"{f:.2f}",
                    fontsize=5,
                    rotation=45,
                    ha="left",
                    va="center",
                )
            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.set_xlim(5, 25)
        return self



from typing import Optional

import numpy as np
import dask.array as da
import h5py

from ..base import Base


class fft(Base):
    def calc(
        self,
        dset: str,
        name: Optional[str] = None,
        override: Optional[bool] = False,
        tslice=slice(None),
        zslice=slice(None),
        yslice=slice(None),
        xslice=slice(None),
        cslice=2,
    ):
        if "dt" not in self.llyr.attrs:
            print("Add the 'dt' attribute before calculating the fft")
            return
        if name is not None and f"{name}/arr" in self.llyr.dsets and not override:
            raise NameError(
                f"'{name}' already exists, you can use 'override=True' to replace it"
            )
        with h5py.File(self.llyr.path, "r") as f:
            arr = da.from_array(f[dset], chunks=(None, None, 16, None, None))
            arr = arr[(tslice, zslice, yslice, xslice, cslice)]
            arr = arr.sum(axis=1)  # sum all z
            arr = da.subtract(arr, arr[0])
            arr = da.subtract(arr, da.average(arr, axis=0)[None, :])
            arr = da.multiply(arr, np.hanning(arr.shape[0])[:, None, None])
            arr = da.swapaxes(arr, 0, -1)
            arr = da.reshape(
                arr, (arr.shape[0] * arr.shape[1], arr.shape[2])
            )  # flatten all the cells
            arr = da.fft.rfft(arr)
            arr = da.absolute(arr)
            arr = da.sum(arr, axis=0)
            arr = arr.compute()

        freqs = np.fft.rfftfreq(self.llyr.h5.shape(dset)[0], self.llyr.dt)

        if name is not None:
            self.llyr.add_dset(arr, f"{name}/arr", override)
            self.llyr.add_dset(freqs, f"{name}/freqs", override)



import matplotlib.pyplot as plt

from ..base import Base


class snapshot_png(Base):
    def plot(self, image: str):
        arr = self.llyr[f"snapshots/{image}"][:]
        fig, ax = plt.subplots(1, 1, figsize=(4, 2))
        ax.imshow(
            arr,
            origin="lower",
            extent=[
                0,
                arr.shape[0] * self.llyr.dy * 1e9,
                0,
                arr.shape[1] * self.llyr.dx * 1e9,
            ],
        )
        ax.set_ylabel(r"$y$ (nm)")
        ax.set_xlabel(r"$x$ (nm)")
        fig.tight_layout()
        return self


import matplotlib.pyplot as plt

from ..base import Base


class imshow(Base):
    def plot(self, dset: str, zero: bool = True, t: int = -1, c: int = 2, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(3, 3), dpi=200)
        else:
            fig = ax.figure
        if zero:
            arr = self.llyr[dset][[0, t], 0, :, :, c]
            arr = arr[1] - arr[0]
        else:
            arr = self.llyr[dset][t, 0, :, :, c]
        amin, amax = arr.min(), arr.max()
        if amin < 0 < amax:
            cmap = "cmo.balance"
            vmm = max((-amin, amax))
            vmin, vmax = -vmm, vmm
        else:
            cmap = "cmo.amp"
            vmin, vmax = amin, amax
        ax.imshow(
            arr,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            extent=[
                0,
                arr.shape[1] * self.llyr.dx * 1e9,
                0,
                arr.shape[0] * self.llyr.dy * 1e9,
            ],
        )
        ax.set(
            title=self.llyr.name,
            xlabel="x (nm)",
            ylabel="y (nm)",
        )
        fig.colorbar(ax.get_images()[0], ax=ax)

        return self



from collections import namedtuple

import matplotlib.pyplot as plt
import peakutils
import numpy as np

from ..base import Base


class report(Base):
    def plot(
        self,
        dset="m",
        thres=0.01,
        min_dist=5,
        tmin=None,
        tmax=None,
        tstep=1,
        nb_modes=6,
        save=False,
    ):
        def get_spectra() -> dict:
            spectra = {}
            for dset in ["mx", "my", "mz"]:
                y = self.llyr[f"table/{dset}"][slice(tmin, tmax, tstep)]
                ts = self.llyr["table/t"][slice(tmin, tmax, tstep)]
                table_dt = (ts[-1] - ts[0]) / len(ts)
                x = np.fft.rfftfreq(y.shape[0], table_dt * tstep) * 1e-9
                y -= y[0]
                y -= np.average(y)
                y = np.multiply(y, np.hanning(y.shape[0]))
                y = np.fft.rfft(y)
                y = np.abs(y)
                spectra["freqs"] = x
                spectra[dset] = y
            return spectra

        def get_peaks(s):
            Peak = namedtuple("Peak", "idx freq amp")
            all_peaks = {}
            for dset in ["mx", "my", "mz"]:
                idx = peakutils.indexes(s[dset], thres=thres, min_dist=min_dist)
                peak_amp = [s[dset][i] for i in idx]
                freqs = [s["freqs"][i] for i in idx]
                all_peaks[dset] = [
                    Peak(i, f, a) for i, f, a in zip(idx, freqs, peak_amp)
                ]
            no_dup_peaks = []
            all_freqs = []
            for peak_list in all_peaks.values():
                for peak in peak_list:
                    if peak.freq not in all_freqs:  # removes duplicate freqs
                        no_dup_peaks.append(peak)
                        all_freqs.append(peak.freq)
            sorted_peaks = sorted(no_dup_peaks, key=lambda x: x[2], reverse=True)[
                :nb_modes
            ]
            return sorted_peaks, all_peaks

        def get_modes(peaks):
            modes = []
            Mode = namedtuple("Mode", "idx freq amp mx my mz")
            ModeComp = namedtuple("ModeArr", "abs ang alpha")
            for peak in peaks:
                modes_comps = []
                for comp in [0, 1, 2]:
                    arr = self.llyr.calc.mode(dset, peak.freq, comp)
                    arr_abs = np.abs(arr)
                    arr_ang = np.angle(arr)
                    arr_alpha = arr_abs / arr_abs.max()
                    modes_comps.append(ModeComp(arr_abs, arr_ang, arr_alpha))

                modes.append(
                    Mode(
                        peak.idx,
                        peak.freq,
                        peak.amp,
                        modes_comps[0],
                        modes_comps[1],
                        modes_comps[2],
                    )
                )
            return modes

        def plot_spectra(gs_main, spectra, all_peaks):
            gs_spectra = gs_main[0].subgridspec(1, 3, wspace=0, hspace=0)
            axes = gs_spectra.subplots()
            axes[1].set_title(self.llyr.name)
            for c, ax in zip(["mx", "my", "mz"], axes):
                peaks = all_peaks[c]
                ax.plot(spectra["freqs"], spectra[c])
                for peak in peaks:
                    ax.text(
                        peak.freq,
                        peak.amp + 0.03 * max(spectra[c]),
                        f"{peak.freq:.2f}",
                        fontsize=5,
                        rotation=90,
                        ha="center",
                        va="bottom",
                    )
                ax.spines["top"].set_visible(False)
                ax.spines["right"].set_visible(False)

        def plot_modes(gs_main, modes):
            nb_rows = int(np.ceil(len(modes) / 3))
            gs_modes = gs_main[1:].subgridspec(nb_rows, 3, wspace=0, hspace=0)
            extent = [
                0,
                modes[0].my.abs.shape[0] * self.llyr.dy * 1e9,
                0,
                modes[0].mx.abs.shape[0] * self.llyr.dx * 1e9,
            ]
            mode_list_max = np.max([np.max(mode.amp) for mode in modes])
            for i, mode in enumerate(modes):
                gs_mode = gs_modes[i].subgridspec(3, 3, wspace=0, hspace=0)
                axes = gs_mode.subplots()
                axes[0, 1].set_title(
                    f"{mode.freq:.2f} GHz (A={mode.amp:.1e})", fontsize=9
                )
                for c in range(3):
                    axes[0, c].imshow(
                        mode[c + 3].abs,
                        cmap="inferno",
                        vmin=0,
                        vmax=mode_list_max,
                        extent=extent,
                        interpolation="None",
                        aspect="equal",
                    )
                    axes[1, c].imshow(
                        mode[c + 3].ang,
                        aspect="equal",
                        cmap="hsv",
                        vmin=-np.pi,
                        vmax=np.pi,
                        interpolation="None",
                        extent=extent,
                    )
                    axes[2, c].imshow(
                        mode[c + 3].ang,
                        aspect="equal",
                        alpha=mode[c + 3].alpha,
                        cmap="hsv",
                        vmin=-np.pi,
                        vmax=np.pi,
                        interpolation="None",
                        extent=extent,
                    )
                for ax in axes.flatten():
                    ax.set(xticks=[], yticks=[])

        spectra = get_spectra()
        sorted_peaks, all_peaks = get_peaks(spectra)
        modes = get_modes(sorted_peaks)

        fig = plt.figure(constrained_layout=True, figsize=(7, 5))
        gs_main = fig.add_gridspec(4, 1)
        plot_spectra(gs_main, spectra, all_peaks)
        plot_modes(gs_main, modes)

        if save:
            fig.savefig(self.llyr.path.replace(".h5", ".pdf"), dpi=100)



from typing import Optional, Dict
import os
import re
import glob
import shutil
import time

import numpy as np
import imageio

# from ._utils import get_config
from .ovf import get_ovf_parms


class Make:
    def __init__(self, llyr):
        self.llyr = llyr
        self.skip_ovf: bool
        self.ts: int
        self.override: bool
        self.out_path: str
        self.mx3_path: str
        self.logs_path: str
        self.dset_prefixes: dict

    def make(
        self,
        load_path: Optional[str] = None,
        tmax: Optional[int] = None,
        override: bool = False,
        delete_out: bool = False,
        delete_mx3: bool = False,
        skip_ovf: bool = False,
    ):
        if load_path is None:
            load_path = self.llyr.path.replace(".h5", ".out")
        self.skip_ovf = skip_ovf
        self.ts = 0
        self.override = override
        self.llyr.h5.create_h5(override)
        self.llyr.h5.add_attr("version", "0.1.5")
        self.add_paths(load_path)
        self.add_times()
        self.add_step_size()
        self.add_mx3()
        self.add_snapshots()
        self.add_logs()
        self.add_table()
        if not skip_ovf:
            self.add_dset_prefixes()
            for prefix, name in self.dset_prefixes.items():
                self.make_dset(prefix, name=name, tmax=tmax)
        if delete_out:
            shutil.rmtree(self.out_path)
        if self.mx3_path != "" and delete_mx3:
            os.remove(self.mx3_path)

    def add_paths(self, load_path: str):
        """Cleans the input string and return the path for .out folder and .mx3 file"""
        load_path = load_path.replace(".mx3", "").replace(".out", "").replace(".h5", "")

        if os.path.exists(f"{load_path}.out"):
            self.out_path = f"{load_path}.out"
        else:
            raise NameError(f"{load_path}.out not found")

        if os.path.exists(f"{load_path}.mx3"):
            self.mx3_path = f"{load_path}.mx3"
        else:
            self.mx3_path = ""

        if os.path.exists(f"{self.out_path}/slurm.logs"):
            self.logs_path = f"{self.out_path}/slurm.logs"
        elif os.path.exists(f"{self.out_path}/log.txt"):
            self.logs_path = f"{self.out_path}/log.txt"
        else:
            self.logs_path = ""

    def add_times(self):
        start_file_path = f"{self.out_path}/gui"
        stop_file_path = f"{self.out_path}/log.txt"
        if os.path.exists(start_file_path):
            self.llyr.h5.add_attr(
                "start_time",
                time.asctime(time.localtime(os.stat(start_file_path).st_mtime)),
            )
        if os.path.exists(stop_file_path):
            self.llyr.h5.add_attr(
                "stop_time",
                time.asctime(time.localtime(os.stat(stop_file_path).st_mtime)),
            )

    def add_step_size(self):
        # load one file to initialize the h5 dataset with the correct shape
        ovf_paths = glob.glob(f"{self.out_path}/*.ovf")
        if len(ovf_paths) == 0:
            return
        else:
            ovf_path = ovf_paths[-2]
        ovf_parms = get_ovf_parms(ovf_path)
        for key in ["dx", "dy", "dz"]:
            if key not in self.llyr.attrs:
                self.llyr.h5.add_attr(key, ovf_parms[key])

    def add_mx3(self):
        """Adds the mx3 file to the f.attrs"""
        if self.mx3_path != "":
            with open(self.mx3_path, "r") as mx3:
                self.llyr.h5.add_attr("mx3", mx3.read())
        else:
            print("mx3 file not found")

    def add_snapshots(self):
        snapshots_paths = glob.glob(f"{self.out_path}/*.png") + glob.glob(
            f"{self.out_path}/*.jpg"
        )
        for p in snapshots_paths:
            snapshot = imageio.imread(p)
            name = p.split("/")[-1].replace(".png", "")
            self.llyr.h5.add_dset(snapshot, f"snapshots/{name}", override=self.override)

    def add_logs(self):
        """Adds the logs file to the f.attrs"""
        if os.path.isfile(self.logs_path):
            with open(self.logs_path, "r") as logs:
                self.llyr.h5.add_attr("logs", logs.read())
        else:
            print("logs not found")

    def add_table(self, dset_name: str = "table"):
        """Adds a the mumax table.txt file as a dataset"""
        table_path = f"{self.out_path}/table.txt"
        if os.path.isfile(table_path):
            with open(table_path, "r") as table:
                header = table.readline()
                data = np.loadtxt(table).T
            # Add dt
            if self.skip_ovf:
                times = data[0]
                dt = (times[-1] - times[0]) / (len(times) - 1)
                self.llyr.h5.add_attr("dt", dt)
            # add table data
            clean_header = [
                i.split(" (")[0].replace("# ", "") for i in header.split("\t")
            ]
            for i, h in enumerate(clean_header):
                self.llyr.h5.add_dset(
                    data[i], f"{dset_name}/{h}", override=self.override
                )

    def add_dset_prefixes(self):
        """From the .out folder, get the list of prefixes, each will correspond to a different dataset"""
        paths = glob.glob(f"{self.out_path}/*.ovf")
        prefixes = list(
            {re.sub(r"_?[\d.]*.ovf", "", path.split("/")[-1]) for path in paths}
        )
        names = {}
        # prefix_to_name = get_config()
        prefix_to_name: Dict[str, str] = dict()
        for prefix in prefixes:
            names[prefix] = prefix
            for pattern, name in prefix_to_name.items():
                if re.findall(pattern, prefix.lower()):
                    names[prefix] = name
                    break
        self.dset_prefixes = names

    def make_dset(
        self,
        prefix: str,
        name: str,
        tmax: Optional[int] = None,
    ):
        """Creates a dataset from an input .out folder path and a prefix (i.e. "m00")"""
        ovf_paths = sorted(glob.glob(f"{self.out_path}/{prefix}*.ovf"))[:tmax]
        # this is to calculate dt
        if self.ts < len(ovf_paths):
            self.ts = len(ovf_paths)

        ovf_parms = get_ovf_parms(ovf_paths[-1])
        if "dt" not in self.llyr.dsets and len(ovf_paths) > 2:
            t0 = get_ovf_parms(ovf_paths[0])["t"]
            tn = ovf_parms["t"]
            dt = (tn - t0) / len(ovf_paths)
            self.llyr.h5.add_attr("dt", dt)
        dset_shape = (len(ovf_paths),) + ovf_parms["shape"]
        self.llyr.h5.load_dset(name, dset_shape, ovf_paths)



from .anim import anim
from .fft_tb import fft_tb
from .imshow import imshow
from .modes import modes
from .snapshot import snapshot
from .snapshot_png import snapshot_png
from .report import report


class Plot:
    def __init__(self, llyr):
        self.anim = anim(llyr).plot
        self.fft_tb = fft_tb(llyr).plot
        self.imshow = imshow(llyr).plot
        self.modes = modes(llyr).plot
        self.snapshot = snapshot(llyr).plot
        self.snapshot_png = snapshot_png(llyr).plot
        self.report = report(llyr).plot



import colorsys

import matplotlib.pyplot as plt
import numpy as np

from .._utils import hsl2rgb
from ..base import Base


class snapshot(Base):
    def plot(self, dset: str, z: int = 0, t: int = -1, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(3, 3), dpi=200)
        else:
            fig = ax.figure
        arr = self.llyr[dset][t, z, :, :, :]
        arr = np.ma.masked_equal(arr, 0)
        u = arr[:, :, 0]
        v = arr[:, :, 1]
        z = arr[:, :, 2]

        alphas = -np.abs(z) + 1
        hsl = np.ones((u.shape[0], u.shape[1], 3))
        hsl[:, :, 0] = np.angle(u + 1j * v) / np.pi / 2 + 0.5  # normalization
        hsl[:, :, 1] = np.sqrt(u ** 2 + v ** 2 + z ** 2)
        hsl[:, :, 2] = (z + 1) / 2
        rgb = hsl2rgb(hsl)
        stepx = max(int(u.shape[1] / 60), 1)
        stepy = max(int(u.shape[0] / 60), 1)
        scale = 1 / max(stepx, stepy)
        x, y = np.meshgrid(
            np.arange(0, u.shape[1], stepx) * self.llyr.dx * 1e9,
            np.arange(0, u.shape[0], stepy) * self.llyr.dy * 1e9,
        )
        antidots = np.ma.masked_not_equal(self.llyr[dset][0, 0, :, :, 2], 0)
        ax.quiver(
            x,
            y,
            u[::stepy, ::stepx],
            v[::stepy, ::stepx],
            alpha=alphas[::stepy, ::stepx],
            angles="xy",
            scale_units="xy",
            scale=scale,
        )

        ax.imshow(
            rgb,
            interpolation="None",
            origin="lower",
            cmap="hsv",
            vmin=-np.pi,
            vmax=np.pi,
            extent=[
                0,
                arr.shape[1] * self.llyr.dx * 1e9,
                0,
                arr.shape[0] * self.llyr.dy * 1e9,
            ],
        )
        ax.imshow(
            antidots,
            interpolation="None",
            origin="lower",
            cmap="Set1_r",
            extent=[
                0,
                arr.shape[1] * self.llyr.dx * 1e9,
                0,
                arr.shape[0] * self.llyr.dy * 1e9,
            ],
        )
        ax.set(title=self.llyr.name, xlabel="x (nm)", ylabel="y (nm)")
        L, H = np.mgrid[0 : 1 : arr.shape[1] * 1j, 0:1:20j]  # type: ignore
        S = np.ones_like(L)
        rgb = hsl2rgb(np.dstack((H, S, L)))
        fig.tight_layout()
        self.add_radial_phase_colormap(ax)
        return self

    def add_radial_phase_colormap(self, ax, rec=None):
        def func1(hsl):
            return np.array(colorsys.hls_to_rgb(hsl[0] / (2 * np.pi), hsl[1], hsl[2]))

        if rec is None:
            rec = [0.85, 0.85, 0.15, 0.15]
        cax = plt.axes(rec, polar=True)
        p1, p2 = ax.get_position(), cax.get_position()
        cax.set_position([p1.x1 - p2.width, p1.y1 - p2.height, p2.width, p2.height])

        theta = np.linspace(0, 2 * np.pi, 360)
        r = np.arange(0, 100, 1)
        hls = np.ones((theta.size * r.size, 3))

        hls[:, 0] = np.tile(theta, r.size)
        white_pad = 20
        black_pad = 10
        hls[: white_pad * theta.size, 1] = 1
        hls[-black_pad * theta.size :, 1] = 0
        hls[white_pad * theta.size : -black_pad * theta.size, 1] = np.repeat(
            np.linspace(1, 0, r.size - white_pad - black_pad), theta.size
        )
        rgb = np.apply_along_axis(func1, 1, hls)
        cax.pcolormesh(
            theta,
            r,
            np.zeros((r.size, theta.size)),
            color=rgb,
            shading="nearest",
        )
        cax.spines["polar"].set_visible(False)
        cax.set(yticks=[], xticks=[])
        up_symbol = dict(x=0.5, y=0.5, name=r"$\bigotimes$")
        down_symbol = dict(x=0.1, y=0.5, name=r"$\bigodot$")
        for s in [up_symbol, down_symbol]:
            cax.text(
                s["x"],
                s["y"],
                s["name"],
                color="#3b5bff",
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=5,
                transform=cax.transAxes,
            )



import os
import configparser
import glob

import numpy as np
from matplotlib import pyplot as plt
import matplotlib as mpl
from appdirs import user_config_dir


def get_config() -> configparser.SectionProxy:
    config_dir = user_config_dir("llyr")
    if os.path.isfile("llyr.ini"):
        config_path = "llyr.ini"
    elif os.path.isfile(f"{config_dir}/llyr.ini"):
        config_path = f"{config_dir}/llyr.ini"
    config = configparser.ConfigParser()
    config.read(config_path)
    return config["llyr"]


def hsl2rgb(hsl):
    h = hsl[..., 0] * 360
    s = hsl[..., 1]
    l = hsl[..., 2]

    rgb = np.zeros_like(hsl)
    for i, n in enumerate([0, 8, 4]):
        k = (n + h / 30) % 12
        a = s * np.minimum(l, 1 - l)
        k = np.minimum(k - 3, 9 - k)
        k = np.clip(k, -1, 1)
        rgb[..., i] = l - a * k
    rgb = np.clip(rgb, 0, 1)
    return rgb


def clean_glob_names(ps):
    ps = sorted([x.split("/")[-1].split(".")[0] for x in ps])
    pre_sub = ""
    for i in range(20):
        q = [pre_sub + ps[0][i] == p[: i + 1] for p in ps]
        if not all(q):
            break
        pre_sub += ps[0][i]
    post_sub = ""
    for i in range(-1, -20, -1):
        q = [ps[0][i] + post_sub == p[i:] for p in ps]
        if not all(q):
            break
        post_sub = ps[0][i] + post_sub
    ps = [p.replace(pre_sub, "").replace(post_sub, "") for p in ps]
    return pre_sub, post_sub, ps


def cspectra_b(Llyr):
    def cspectra(ps, norm=None):
        cmaps = []
        for a, b, c in zip((1, 0, 0), (0, 1, 0), (0, 0, 1)):
            N = 256
            vals = np.ones((N, 4))
            vals[:, 0] = np.linspace(1, a, N)
            vals[:, 1] = np.linspace(1, b, N)
            vals[:, 2] = np.linspace(1, c, N)
            vals[:, 3] = np.linspace(0, 1, N)
            cmaps.append(mpl.colors.ListedColormap(vals))
        paths = glob.glob(f"{ps}/*.h5")[:17]
        fig, ax = plt.subplots(1, 1, figsize=(5, 5), sharex=True, sharey=True)
        for c, cmap in zip(["mx", "mz"], [cmaps[0], cmaps[2]]):
            names = []
            arr = []
            for p in paths:
                m = Llyr(p)
                names.append(m.name)
                x, y = m.fft_tb(c, tmax=None, normalize=True)
                x, y = x[5:], y[5:]
                arr.append(y)
            arr = np.array(arr).T
            # norm=mpl.colors.SymLogNorm(linthresh=0.2)

            ax.imshow(
                arr,
                aspect="auto",
                origin="lower",
                interpolation="nearest",
                extent=[0, arr.shape[1] * 2, x.min(), x.max()],
                cmap=cmap,
                norm=norm,
            )
        ax.legend(
            handles=[
                mpl.patches.Patch(color="red", label="mx"),
                mpl.patches.Patch(color="blue", label="mz"),
            ],
            fontsize=5,
        )
        _, _, ps = clean_glob_names(paths)
        ax.set_ylim(0, 15)
        ax.set_xticks(np.arange(1, arr.shape[1] * 2, 2))
        ax.set_xticklabels(ps)
        ax.set_xlabel("Ring Width (nm)")
        ax.set_ylabel("Frequency (GHz)")
        fig.tight_layout(h_pad=0.4, w_pad=0.2)
        return fig, ax

    return cspectra



import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from .._utils import hsl2rgb

from ..base import Base


class anim(Base):
    def plot(self, dset: str, z: int = 0, tmin: int = 0, tmax: int = -1):
        arr = self.llyr[dset][tmin:tmax, z, :, :, :]
        # arr -= arr[0]
        fig, ax = plt.subplots(1, 1, figsize=(3, 3), dpi=200)
        # Arrays
        arr = np.ma.masked_equal(arr, 0)
        u, v, z = arr[..., 0], arr[..., 1], arr[..., 2]
        alphas = -np.abs(z) + 1
        hsl = np.ones((u.shape[0], u.shape[1], u.shape[2], 3))
        hsl[..., 0] = np.angle(u + 1j * v) / np.pi / 2 + 0.5  # normalization
        hsl[..., 1] = np.sqrt(u ** 2 + v ** 2 + z ** 2)
        hsl[..., 2] = (z + 1) / 2
        rgb = hsl2rgb(hsl)
        stepx = max(int(u.shape[2] / 60), 1)
        stepy = max(int(u.shape[1] / 60), 1)
        scale = 1 / max(stepx, stepy)
        x, y = np.meshgrid(
            np.arange(0, u.shape[2], stepx) * self.llyr.dx * 1e9,
            np.arange(0, u.shape[1], stepy) * self.llyr.dy * 1e9,
        )
        antidots = np.ma.masked_not_equal(self.llyr["m"][0, 0, :, :, 2], 0)
        extent = [
            0,
            arr.shape[2] * self.llyr.dx * 1e9,
            0,
            arr.shape[1] * self.llyr.dy * 1e9,
        ]
        t = 0
        Q = ax.quiver(
            x,
            y,
            u[t, ::stepy, ::stepx],
            v[t, ::stepy, ::stepx],
            alpha=alphas[t, ::stepy, ::stepx],
            angles="xy",
            scale_units="xy",
            scale=scale,
        )

        ax.imshow(
            rgb[t],
            interpolation="None",
            origin="lower",
            cmap="hsv",
            vmin=-np.pi,
            vmax=np.pi,
            extent=extent,
        )
        ax.imshow(
            antidots, interpolation="None", origin="lower", cmap="Set1_r", extent=extent
        )
        ax.get_images()[0].set_data(rgb[t + 1])
        ax.set(title=self.llyr.name, xlabel="x (nm)", ylabel="y (nm)")
        fig.tight_layout()

        def run(t):
            ax.get_images()[0].set_data(rgb[t])
            Q.set_UVC(u[t, ::stepy, ::stepx], v[t, ::stepy, ::stepx])
            Q.set_alpha(alphas[t, ::stepy, ::stepx])
            ax.set_title(f"t={t}")
            return ax

        ani = mpl.animation.FuncAnimation(
            fig, run, interval=1, frames=np.arange(1, arr.shape[0], dtype="int")
        )
        anim_save_path = self.llyr.path.replace(".h5", ".mp4")
        ani.save(
            anim_save_path,
            writer="ffmpeg",
            fps=25,
            dpi=300,
            extra_args=["-vcodec", "h264", "-pix_fmt", "yuv420p"],
        )
        print(f"Saved at: {anim_save_path}")
        plt.close()
        return self


import numpy as np

from ..base import Base


class sk_number(Base):
    def calc(self, dset: str, z: int = 0, t: int = 0):
        spin_grid = self.llyr[dset][t, z, :, :, :]
        spin_pad = np.pad(
            spin_grid, ((1, 1), (1, 1), (0, 0)), mode="constant", constant_values=0.0
        )
        sk_nb = (
            np.cross(
                spin_pad[2:, 1:-1, :],  # s(i+1,j)
                spin_pad[1:-1, 2:, :],  # s(i,j+1)
                axis=2,
            )
            + np.cross(
                spin_pad[:-2, 1:-1, :],  # s(i-1,j)
                spin_pad[1:-1, :-2, :],  # s(i,j-1)
                axis=2,
            )
            - np.cross(
                spin_pad[:-2, 1:-1, :],  # s(i-1,j)
                spin_pad[1:-1, 2:, :],  # s(i,j+1)
                axis=2,
            )
            - np.cross(
                spin_pad[2:, 1:-1, :],  # s(i+1,j)
                spin_pad[1:-1, :-2, :],  # s(i,j-1)
                axis=2,
            )
        )
        sk_nb = -1 * np.einsum("ijk,ijk->ij", spin_grid, sk_nb) / (16 * np.pi)
        return np.sum(sk_nb.flatten())
